/* KallistiOS 1.1.3

   pcx_texture.c
   (c)2000-2001 Dan Potter

   PCX texture loader
*/

static char id[] = "KOS $Id: pcx_texture.c,v 1.1 2001/08/27 05:50:26 bard Exp $";

#include <kos.h>

/* Loads a PCX file into texture RAM, potentially twiddling it and/or
   converting it to an alpha mask. */
int pcx_load_texture(const char *fn, int twiddle, int alpha,
		uint32 *txr_out, int *w, int *h) {
	Image	*img;
	uint32	txr;

	/* Load the image */
	img = pcx_load(fn);
	if (img == NULL)
		return -1;

	printf("image is %dx%d\n", img->width, img->height);
	printf("w is %08x, h is %08x\n", w, h);

	/* Allocate texture RAM for it */
	txr = ta_txr_allocate(img->width * img->height * 2);

	/* If requested, convert it to an alpha mask */
	if (alpha)
		txr_to_alpha(img->pixel_data, img->width, img->height);

	/* Now either twiddle it or copy it to texture ram */
	if (twiddle)
		txr_twiddle_copy(img->pixel_data, img->width, img->height,
			txr, img->width, img->height, 0);
	else
		ta_txr_load(txr, img->pixel_data, img->width*img->height*2);

	if (txr_out != NULL) *txr_out = txr;
	if (w != NULL) *w = img->width;
	if (h != NULL) *h = img->height;

	pcx_free(img);

	return 0;
}

